/*
 * Copyright © 2008 Étienne Bersac <bersace03@gmail.com>
 * 
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of version 2 of the GNU General Public License as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02110 USA
 *
 */

#include <glib.h>
#include <hal/libhal.h>
#include <dbus/dbus-glib-lowlevel.h>
#include <sane/sane.h>
#include "addon.h"
#include "scanner.h"

/*
 * Add a timeout function for polling @scanner.
 */
static void
addon_add_poll(Scanner *scanner)
{
	debug("Start polling");

	/* how to stop polling for saving power? */
	g_timeout_add(500,
		      (GSourceFunc) scanner_poll,
		      scanner);
}

/*
 * Handle org.freedesktop.Hal.Device.Scanner method call.
 */
static DBusHandlerResult
addon_filter(DBusConnection *dbus,
	     DBusMessage *msg,
	     void* user_data)
{
	Scanner *scanner = (Scanner*) user_data;
	const gchar *member = dbus_message_get_member(msg);

	if (g_str_equal(member, "ReleaseDevice")) {
		/* closing the scanner autostop polling */
		/* TODO: return a boolean? Run async? */
		scanner_close(scanner);
		return DBUS_HANDLER_RESULT_HANDLED;
	}
	else if (g_str_equal(member, "ClaimDevice")) {
		/* ensure no polling */
		scanner_close(scanner);
		g_usleep(500);
		/* reopen the scan and add poll */
		if (scanner_open(scanner))
			addon_add_poll(scanner);
		return DBUS_HANDLER_RESULT_HANDLED;
	}

	return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
}

gint
main (gint argc,
      gchar **argv)
{
	Scanner *scanner;
	LibHalContext *hal;
	DBusConnection *dbus;
	const gchar *udi;
	GMainLoop *main_loop;
	DBusError err;
	SANE_Int version;
	SANE_Status st;
	
	openlog(g_get_prgname(), LOG_PERROR, LOG_USER);

	udi = g_getenv ("UDI");

	if (udi == NULL)
		error ("No device specified");

	dbus_error_init (&err);
	if ((hal = libhal_ctx_init_direct (&err)) == NULL)
		error ("Cannot connect to hald");

	dbus_error_init (&err);
	if (!libhal_device_addon_is_ready (hal, udi, &err))
		return -4;

	/* init SANE */
	st = sane_init(&version, NULL);
	if (st != SANE_STATUS_GOOD)
		error("Failed to init SANE: %s", sane_strstatus(st));

	/* open device */
	if ((scanner = scanner_new(hal, udi)) == NULL)
		error("Unable to poll %s", udi);
	
	dbus = libhal_ctx_get_dbus_connection (hal);
	dbus_connection_setup_with_g_main (dbus, NULL);
	dbus_connection_set_exit_on_disconnect (dbus, 0);

	dbus_error_init(&err);
	libhal_device_claim_interface(hal, udi,
				      HAL_SCANNER_IFACE,
				      /* Release Device method */
				      "<method name=\"ReleaseDevice\"/>"
				      /* claim device method */
				      "<method name=\"ClaimDevice\"/>"
				      /* Event signal */
				      "<signal name=\"Event\">\n"
				      "<arg name=\"option_name\" type=\"s\"/>"
				      "<arg name=\"value\" type=\"v\"/>"
				      "</signal>",
				      &err);

	dbus_connection_add_filter(dbus, addon_filter, scanner, NULL);

	dbus_error_init(&err);
	gchar *rule = g_strdup_printf("interface='" HAL_SCANNER_IFACE "'"
				      ",path='%s'", udi);
	dbus_bus_add_match(dbus, rule, &err);
	g_free(rule);

	addon_add_poll(scanner);

	main_loop = g_main_loop_new (NULL, FALSE);

	g_main_loop_run (main_loop);

	/* is that dead code? */
	closelog();

	return 0;
}
