/*
 * Copyright © 2008 Étienne Bersac <bersace03@gmail.com>
 * 
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of version 2 of the GNU General Public License as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02110 USA
 *
 */

#include "addon.h"
#include "scanner.h"

/* PRIVATE TYPES */

typedef union _Value Value;
union _Value
{
	SANE_Bool	boolean;
	SANE_Word	word;
	SANE_String	string;
};

typedef struct _Option Option;
struct _Option
{
	Scanner		*scanner;
	SANE_Int	n;
	const SANE_Option_Descriptor	*desc;
	guchar		dbus_type;
	Value		*value;
};


/* OPTION HANDLING */

/*
 * Free existing stored option value.
 */
static void
option_unset(Option *self)
{
	if (!self->value)
		return;
	
	switch(self->desc->type) {
	case SANE_TYPE_STRING:
		g_free(self->value->string);
		break;
	default:
		break;
	}
	g_free(self->value);
	self->value = NULL;
}

/*
 * unset option and destroy it.
 */
static gpointer
option_destroy(Option *self)
{
	option_unset(self);
	g_free(self);
	return NULL;
}



/*
 * Update stored value and return wether the option has changed. See
 * sane_control_option().
 */
static gboolean
option_update(Option *self)
{
	Value *value;
	SANE_Word word;
	SANE_String string = NULL;
	gpointer data = NULL;
	SANE_Status st;
	gboolean changed = FALSE;


	switch(self->desc->type) {
	case SANE_TYPE_STRING:
		string = g_new0(SANE_Char, self->desc->size);
		data = string;
		break;
	case SANE_TYPE_INT:
	case SANE_TYPE_FIXED:
	case SANE_TYPE_BOOL:
		data = &word;
		break;
	default:
		return changed;
		break;
	}

	st = sane_control_option(self->scanner->handle,
				 self->n,
				 SANE_ACTION_GET_VALUE,
				 data, NULL);

	if (st != SANE_STATUS_GOOD) {
		warn("Error while reading option %i:%s: %s",
		     self->n, self->desc->name, sane_strstatus(st));
		return changed;
	}

	value = g_new0(Value, 1);

	switch(self->desc->type) {
	case SANE_TYPE_STRING:
		value->string = string;
		break;
	case SANE_TYPE_INT:
	case SANE_TYPE_FIXED:
	case SANE_TYPE_BOOL:
		value->word = word;
		break;
	default:
		return changed;
		break;
	}

	if (self->value) {
		switch(self->desc->type) {
		case SANE_TYPE_STRING:
			changed = !g_str_equal(self->value->string, value->string);
			break;
		case SANE_TYPE_INT:
		case SANE_TYPE_FIXED:
			changed = value->word != self->value->word;
			break;
		case SANE_TYPE_BOOL:
			changed = (value->boolean && !self->value->boolean)
				|| (!value->boolean && self->value->boolean);
			break;
		default:
			break;
		}

		if (changed) {
			debug("%s changed to %i",
			      self->desc->name,
			      self->value->word);
		}
		option_unset(self);
	}
	self->value = value;

	return changed;
}

static guchar
option_get_dbus_type(Option *self)
{
	switch(self->desc->type) {
	case SANE_TYPE_STRING:
		return DBUS_TYPE_STRING;
		break;
	case SANE_TYPE_INT:
		return DBUS_TYPE_INT32;
		break;
	case SANE_TYPE_FIXED:
		return DBUS_TYPE_DOUBLE;
		break;
	case SANE_TYPE_BOOL:
		return DBUS_TYPE_BOOLEAN;
		break;
	default:
		return DBUS_TYPE_INVALID;
		break;
	}
}


/*
 * Instanciate a new option. groupe may be NULL. If desc is NULL, desc
 * will be retrieved.
 */
static Option*
option_new (Scanner *scanner,
	    SANE_Int n)
{
	Option *self = g_new0(Option, 1);

	self->scanner	= scanner;
	self->n		= n;
	self->desc	= sane_get_option_descriptor(scanner->handle, n);
	if (self->desc->cap & SANE_CAP_AUTOMATIC) {
		sane_control_option(scanner->handle, n,
				    SANE_ACTION_SET_AUTO, NULL, NULL);
	}
		
	/* ignore software settable options, buttons and groups */
	if (n == 0 || (self->desc->cap & SANE_CAP_HARD_SELECT
		       && self->desc->type != SANE_TYPE_BUTTON
		       && self->desc->type != SANE_TYPE_GROUP
		       && SANE_OPTION_IS_ACTIVE(self->desc->cap))) {
		option_update(self);
		return self;
	}
	else
		return option_destroy(self);
}


/* SCANNER HANDLING */

/*
 * compute the SANE device name
 */
static gchar*
scanner_compute_sane_name(LibHalContext *hal,
			  const gchar *udi)
{
	gchar **backend, *type, *id, *name;
	gint bus, dev;
	DBusError err;

	dbus_error_init(&err);
	backend = libhal_device_get_property_strlist(hal, udi,
						     "scanner.sane.backend",
						     &err);
	if (!backend)
		error("device not supported by SANE");

	dbus_error_init(&err);
	type = libhal_device_get_property_string(hal, udi,
						 "info.bus", &err);

	/* generate libusb:%bus:%dev id. This will fail for non libusb
	   backend. */
	if (g_str_equal(type, "usb")) {
		dbus_error_init(&err);
		bus = libhal_device_get_property_int(hal, udi,
						     "usb.bus_number", &err);
	
		dbus_error_init(&err);
		dev = libhal_device_get_property_int(hal, udi,
						     "usb.linux.device_number",
						     &err);
		id = g_strdup_printf("libusb:%03i:%03i", bus, dev);
	}
	/* generate SCSI id. */
	else if (g_str_equal(type, "scsi")) {
		/* TODO: check if this is the right key. */
		dbus_error_init(&err);
		id = libhal_device_get_property_string(hal, udi,
						       "linux.device_file",
						       &err);
	}
	/* TODO: hpaio, net? */
	else {
		warn("Unable to compute SANE device name for %s",
		     udi);
		return NULL;
	}

	name = g_strdup_printf("%s:%s", backend[0], id);
	g_free(id);

	return name;
}

/*
 *
 */
gpointer
scanner_destroy(Scanner *self)
{
	g_slist_foreach(self->options,
			(GFunc) option_destroy, NULL);
	g_slist_free(self->options);
	scanner_close(self);
	g_free(self->name);
	g_free(self);

	return NULL;
}

gboolean
scanner_open(Scanner *self)
{
	SANE_Status st;

	if (self->opened)
		return TRUE;

	st = sane_open(self->name, &self->handle);
	if (st != SANE_STATUS_GOOD) {
		self->opened = FALSE;
		warn("Unable to open device %s: %s",
		     self->name, sane_strstatus(st));
	}
	else {
		self->opened = TRUE;
		debug("Opened SANE device %s", self->name);
	}

	return self->opened;
}

gboolean
scanner_close(Scanner *self)
{
	if (!self->opened)
		return TRUE;

	sane_close(self->handle);
	self->opened = FALSE;
	debug("Closed SANE device %s", self->name);

	return TRUE;
}

/*
 * Open a new scanner, list its buttons.
 */
Scanner*
scanner_new(LibHalContext *hal,
	    const gchar *udi)
{
	DBusError err;
	Scanner *self = g_new0(Scanner, 1);
	Option *opcount, *option;
	gint n;

	self->hal	= hal;
	self->udi	= udi;
	self->name	= scanner_compute_sane_name(hal, udi);

	if (!scanner_open(self))
		return NULL;

	dbus_error_init(&err);
	libhal_device_set_property_string(self->hal, self->udi,
					  "scanner.sane.name", self->name,
					  &err);

	/* search hard selected option */
	opcount = option_new(self, 0);
	for (n = 1; n < opcount->value->word; n++)
		if ((option = option_new(self, n)) != NULL)
			self->options = g_slist_append(self->options, option);

	if (g_slist_length(self->options) != 0)
		return self;
	else {
		debug("No buttons to monitor");
		return scanner_destroy(self);
	}
}

gboolean
scanner_poll(Scanner *self)
{
	if (!self->opened) {
		debug("Stop polling");
		return FALSE;
	}

	GSList *node;
	DBusConnection *dbus;
	DBusMessage *msg;
	Option *option;

	for (node = self->options; node; node = node->next) {
		if (option_update(node->data)) {
			option = node->data;
			msg = dbus_message_new_signal(self->udi,
						      HAL_SCANNER_IFACE,
						      "Event");
			dbus_message_append_args(msg,
						 DBUS_TYPE_STRING,
						 &option->desc->name,
						 option_get_dbus_type(option),
						 option->value,
						 DBUS_TYPE_INVALID);

			dbus = libhal_ctx_get_dbus_connection(self->hal);
			dbus_connection_send(dbus, msg, NULL);
			dbus_connection_flush(dbus);
			dbus_message_unref(msg);

			/* don't test next button, its useless */
			break;
		}
	}

	return TRUE;
}
